use adw::prelude::*;
use rbw::db::{Entry, EntryData, Uri};
use relm4::prelude::{AsyncComponentParts, SimpleAsyncComponent};
use relm4::{adw, gtk, AsyncComponentSender, RelmWidgetExt};
use relm4_icons::icon_names;

#[derive(PartialEq, Clone)]
pub enum DetailMode {
    Display,
    Edit,
}

#[derive(Debug)]
pub enum DetailMsg {
    Edit,
    Discard,
    ClipboardCopy(String),
}

#[derive(Clone)]
pub struct DetailCard {
    mode: DetailMode,
    id: String,
    name: String,
    cardholder_name: Option<String>,
    number: Option<String>,
    brand: Option<String>,
    exp_month: Option<String>,
    exp_year: Option<String>,
    code: Option<String>,
}

#[relm4::component(async, pub)]
impl SimpleAsyncComponent for DetailCard {
    type Init = Entry;
    type Input = ();
    type Output = DetailMsg;

    view! {
        #[root]
        root_box = &gtk::Box {
            set_orientation: gtk::Orientation::Vertical,
            set_hexpand:true,

            gtk::Label {
                set_text: model.name.as_ref()
            },
            adw::EntryRow {
                set_editable: model.mode == DetailMode::Edit,
                set_title: "Cardholder Name",
                set_text: model.cardholder_name.as_ref().unwrap_or(&String::new()),
                add_suffix = &gtk::Button {
                    set_icon_name: icon_names::COPY_REGULAR,
                    set_tooltip: "Copy Username",
                    connect_clicked[sender, model] => move |_| {
                        sender.output(DetailMsg::ClipboardCopy(model.cardholder_name.clone().unwrap_or_else(String::new))).unwrap();
                    }
                }
            },
            adw::PasswordEntryRow {
                set_editable: model.mode == DetailMode::Edit,
                set_title: "Card Number",
                set_text: model.number.as_ref().unwrap_or(&String::new()),
                add_suffix = &gtk::Button {
                    set_icon_name: icon_names::COPY_REGULAR,
                    set_tooltip: "Copy Username",
                    connect_clicked[sender, model] => move |_| {
                        sender.output(DetailMsg::ClipboardCopy(model.number.clone().unwrap_or_else(String::new))).unwrap();
                    }
                }
            },
            adw::PasswordEntryRow {
                set_editable: model.mode == DetailMode::Edit,
                set_title: "Security Code",
                set_text: model.code.as_ref().unwrap_or(&String::new()),
                add_suffix = &gtk::Button {
                    set_icon_name: icon_names::COPY_REGULAR,
                    set_tooltip: "Copy Username",
                    connect_clicked[sender, model] => move |_| {
                        sender.output(DetailMsg::ClipboardCopy(model.code.clone().unwrap_or_else(String::new))).unwrap();
                    }
                }
            },
            // TODO add icons?
            adw::EntryRow {
                set_editable: model.mode == DetailMode::Edit,
                set_title: "Brand",
                set_text: model.brand.as_ref().unwrap_or(&String::new()),
                add_suffix = &gtk::Button {
                    set_icon_name: icon_names::COPY_REGULAR,
                    set_tooltip: "Copy Username",
                    connect_clicked[sender, model] => move |_| {
                        sender.output(DetailMsg::ClipboardCopy(model.brand.clone().unwrap_or_else(String::new))).unwrap();
                    }
                }
            },
            // TODO format it?
            adw::EntryRow {
                set_editable: model.mode == DetailMode::Edit,
                set_title: "Month of Expiration",
                set_text: model.exp_month.as_ref().unwrap_or(&String::new()),
                add_suffix = &gtk::Button {
                    set_icon_name: icon_names::COPY_REGULAR,
                    set_tooltip: "Copy Username",
                    connect_clicked[sender, model] => move |_| {
                        sender.output(DetailMsg::ClipboardCopy(model.exp_month.clone().unwrap_or_else(String::new))).unwrap();
                    }
                }
            },
            adw::EntryRow {
                set_editable: model.mode == DetailMode::Edit,
                set_title: "Year of Expiration",
                set_text: model.exp_year.as_ref().unwrap_or(&String::new()),
                add_suffix = &gtk::Button {
                    set_icon_name: icon_names::COPY_REGULAR,
                    set_tooltip: "Copy Username",
                    connect_clicked[sender, model] => move |_| {
                        sender.output(DetailMsg::ClipboardCopy(model.exp_year.clone().unwrap_or_else(String::new))).unwrap();
                    }
                }
            },
        }
    }

    async fn init(
        entry_data: Self::Init,
        root: Self::Root,
        sender: AsyncComponentSender<Self>,
    ) -> AsyncComponentParts<Self> {
        let model = match entry_data.data {
            EntryData::Login { .. } => todo!(),
            EntryData::Card {
                cardholder_name,
                number,
                brand,
                exp_month,
                exp_year,
                code,
            } => DetailCard {
                mode: DetailMode::Display,
                id: entry_data.id,
                name: entry_data.name,
                cardholder_name,
                number,
                brand,
                exp_month,
                exp_year,
                code,
            },
            EntryData::Identity { .. } => todo!(),
            EntryData::SecureNote => todo!(),
        };
        let widgets = view_output!();

        AsyncComponentParts { model, widgets }
    }

    async fn update(&mut self, _message: Self::Input, _sender: AsyncComponentSender<Self>) {}
}

#[derive(Clone)]
pub struct DetailLogin {
    mode: DetailMode,
    id: String,
    name: String,
    username: Option<String>,
    password: Option<String>,
    totp: Option<String>,
    uris: Vec<Uri>,
}

#[relm4::component(async, pub)]
impl SimpleAsyncComponent for DetailLogin {
    type Init = Entry;
    type Input = DetailMsg;
    type Output = DetailMsg;

    view! {
        #[root]
        root_box = &gtk::Box {
            set_orientation: gtk::Orientation::Vertical,
            set_hexpand:true,

            gtk::Label {
                set_text: model.name.as_ref()
            },
            adw::EntryRow {
                set_title: "Username",
                set_editable: model.mode == DetailMode::Edit,
                set_text: model.username.as_ref().unwrap_or(&String::new()),
                add_suffix = &gtk::Button {
                    set_icon_name: icon_names::COPY_REGULAR,
                    set_tooltip: "Copy Username",
                    connect_clicked[sender, model] => move |_| {
                        sender.output(DetailMsg::ClipboardCopy(model.username.clone().unwrap_or_else(String::new))).unwrap();
                    }
                }
            },
            adw::PasswordEntryRow {
                set_title: "Password",
                set_editable: model.mode == DetailMode::Edit,
                set_text: model.password.as_ref().unwrap_or(&String::new()),
                add_suffix = &gtk::Button {
                    set_icon_name: icon_names::COPY_REGULAR,
                    set_tooltip: "Copy Password",
                    connect_clicked[sender, model] => move |_| {
                        sender.output(DetailMsg::ClipboardCopy(model.password.clone().unwrap_or_else(String::new))).unwrap();
                    }
                }
            },
            adw::EntryRow {
                set_title: "TOTP",
                set_editable: model.mode == DetailMode::Edit,
                set_visible: model.totp.as_ref().is_some_and(|value| !value.is_empty()),
                set_text: model.totp.as_ref().unwrap_or(&String::new()),
                add_suffix = &gtk::Button {
                    set_icon_name: icon_names::COPY_REGULAR,
                    set_tooltip: "Copy TOTP Token"
                }
            },
            #[name = "uris_anchor"]
            gtk::Box {
                set_orientation: gtk::Orientation::Vertical,
                set_hexpand: true,
            },
            // gtk::Button {
            //     set_label: "Edit",
            //     set_visible: model.mode == DetailMode::Display,
            //     set_icon_name: icon_names::EDIT,
            //     connect_clicked[sender] => move |_| {
            //         sender.output(DetailMsg::Edit).unwrap();
            //     }
            // }
        }
    }

    async fn init(
        entry_data: Self::Init,
        root: Self::Root,
        sender: AsyncComponentSender<Self>,
    ) -> AsyncComponentParts<Self> {
        let model = match entry_data.data {
            EntryData::Login {
                username,
                password,
                totp,
                uris,
            } => DetailLogin {
                mode: DetailMode::Display,
                id: entry_data.id,
                name: entry_data.name,
                username,
                password,
                totp,
                uris,
            },
            EntryData::Card { .. } => todo!(),
            EntryData::Identity { .. } => todo!(),
            EntryData::SecureNote => todo!(),
        };

        let widgets = view_output!();

        for uri in model.uris.iter() {
            widgets.uris_anchor.append(
                &adw::EntryRow::builder()
                    .title("Uri")
                    .text(uri.uri.clone())
                    .hexpand(true)
                    .editable(model.mode == DetailMode::Edit)
                    .build(),
            );
        }

        AsyncComponentParts { model, widgets }
    }

    async fn update(&mut self, message: Self::Input, _sender: AsyncComponentSender<Self>) {
        match message {
            DetailMsg::Edit => {
                self.mode = DetailMode::Edit;
            }
            DetailMsg::Discard => self.mode = DetailMode::Display,
            // We do not need to do anything here because it will be
            // escalated to the parent component
            DetailMsg::ClipboardCopy(_) => {}
        }
    }
}
