use adw::prelude::*;
use components::vault::{VaultCreationSetting, VaultCreationToggleMsg};
use copypasta::{ClipboardContext, ClipboardProvider};
use rbw::api::Client;
use rbw::db::{Db, Entry, EntryData};
use rbw::error::Error;
use rbw::identity::Identity;
use rbw::locked::Keys;
use relm4::abstractions::Toaster;
use relm4::actions::{RelmAction, RelmActionGroup};
use relm4::adw::Toast;
use relm4::component::{AsyncComponent, AsyncComponentParts, AsyncComponentSender};
use relm4::factory::AsyncFactoryVecDeque;
use relm4::gtk::glib;
use relm4::prelude::{AsyncComponentController, AsyncController};
use relm4::{adw, gtk, MessageBroker};
use relm4::{RelmApp, RelmWidgetExt};
use std::clone::Clone;
use std::fmt::Debug;
use std::path::PathBuf;
use std::time::Duration;
use wl_clipboard_rs::copy::{MimeType, Options, Source};

mod components;
mod util;
use crate::components::detail::{DetailCard, DetailLogin, DetailMsg};
use crate::components::list::{ListItem, ListOutput};
use crate::components::vault::VaultCreationModel;
use crate::util::{init_db, insert_vault_setting_to_db};
use util::{
    decrypt, load_active_vault_from_db, load_vault_settings_from_db, save_active_vault_to_db,
};

static DEVICE_ID: &str = "bitritter";

static DIALOG_BROKER: MessageBroker<VaultCreationToggleMsg> = MessageBroker::new();

// TODO check for dead enum members
#[derive(Debug)]
pub enum AppMsg {
    Startup,
    // AddVaultItem,
    SetupVault,
    CreateVault(VaultCreationSetting),
    UnlockActiveVault(String),
    LockActiveVault,
    Sync,
    Login(String),
    SetVaultItemFilter(VaultStackState),
    ShowDetailView(String),
    ClipboardCopy(String),
    Edit,
    Discard,
    SwitchVault(VaultSetting),
}

#[derive(Debug)]
pub enum VaultStackState {
    All,
    Login,
    Card,
    Identity,
    Note,
}

struct AppModel {
    vault_stack_state: VaultStackState,
    vaults: Vec<Vault>,
    active_vault: Option<Vault>,
    active_entry: Option<Entry>,
    items: AsyncFactoryVecDeque<ListItem>,
    toaster: Toaster,
    dialog: AsyncController<VaultCreationModel>,
}

#[derive(PartialEq, Eq, Clone, Debug)]
enum VaultState {
    Locked,
    Unlocked,
}

#[derive(Debug)]
pub struct VaultSetting {
    base_url: String,
    email: String,
    identity_url: Option<String>,
    // client_cert_path: Option<PathBuf>,
}

impl From<VaultCreationSetting> for VaultSetting {
    fn from(value: VaultCreationSetting) -> Self {
        Self {
            base_url: value.base_url.clone(),
            email: value.email.clone(),
            identity_url: value.identity_url.clone(),
        }
    }
}

impl From<Vault> for VaultSetting {
    fn from(value: Vault) -> Self {
        Self {
            base_url: value.base_url.clone(),
            email: value.email.clone(),
            identity_url: Some(value.identity_url.clone()),
        }
    }
}

struct Vault {
    base_url: String,
    identity_url: String,
    client_cert_path: Option<PathBuf>,
    email: String,
    state: VaultState,
    entries: Option<Vec<Entry>>,
    unlocked_keys: Option<Keys>,
    db: Db,
}

impl From<VaultSetting> for Vault {
    fn from(value: VaultSetting) -> Self {
        Vault::new(
            value.base_url,
            value.identity_url.unwrap_or_default(),
            None,
            value.email,
        )
    }
}

impl From<VaultCreationSetting> for Vault {
    fn from(value: VaultCreationSetting) -> Self {
        Vault::new(
            value.base_url,
            value.identity_url.unwrap_or_default(),
            None,
            value.email,
        )
    }
}

impl Debug for Vault {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Vault")
            .field("base_url", &self.base_url)
            .field("identity_url", &self.identity_url)
            .field("email", &self.email)
            .finish_non_exhaustive()
    }
}

impl Clone for Vault {
    fn clone(&self) -> Self {
        Self {
            base_url: self.base_url.clone(),
            identity_url: self.identity_url.clone(),
            client_cert_path: self.client_cert_path.clone(),
            email: self.email.clone(),
            state: VaultState::Locked,
            entries: None,
            unlocked_keys: None,
            db: Db::new(),
        }
    }
}

impl Vault {
    #[must_use]
    pub fn new(
        base_url: String,
        identity_url: String,
        client_cert_path: Option<PathBuf>,
        email: String,
    ) -> Vault {
        Vault {
            base_url,
            identity_url,
            client_cert_path,
            email,
            // keys: None,
            state: VaultState::Locked,
            entries: None,
            unlocked_keys: None,
            db: Db::new(),
        }
    }

    pub fn get_client(&self) -> Client {
        Client::new(
            self.base_url.as_str(),
            self.identity_url.as_str(),
            self.client_cert_path.as_deref(),
        )
    }

    pub fn is_logged_in(&self) -> bool {
        self.db.access_token.is_some()
            && self.db.refresh_token.is_some()
            && self.db.kdf.is_some()
            && self.db.iterations.is_some()
    }

    pub async fn pre_login(&mut self) -> Result<(), Error> {
        let client = self.get_client();
        let (kdf_type, iterations, memory, parallelism) =
            client.prelogin(self.email.as_str()).await?;

        self.db.kdf = Some(kdf_type);
        self.db.iterations = Some(iterations);
        self.db.memory = memory;
        self.db.parallelism = parallelism;
        self.db.save(&self.base_url, &self.email)?;
        Ok(())
    }

    pub async fn login(&mut self, password: &str) -> Result<(), Error> {
        let client = self.get_client();

        let mut password_vec = rbw::locked::Vec::new();
        password_vec.extend(password.bytes());
        let password = rbw::locked::Password::new(password_vec.clone());

        let identity = Identity::new(
            self.email.as_str(),
            &password,
            self.db.kdf.expect("KDF should be set"),
            self.db.iterations.expect("iterations should be set"),
            self.db.memory,
            self.db.parallelism,
        )?;
        let (access_token, refresh_token, _key3) = client
            .login(
                self.email.as_str(),
                DEVICE_ID,
                &identity.master_password_hash,
                None,
                None,
            )
            .await?;

        self.db.access_token = Some(access_token);
        self.db.refresh_token = Some(refresh_token);
        self.db.save(&self.base_url, &self.email)?;
        Ok(())
    }

    pub fn unlock(&mut self, password: &str) -> Result<(), Error> {
        let mut password_vec = rbw::locked::Vec::new();
        password_vec.extend(password.bytes());

        let password = rbw::locked::Password::new(password_vec.clone());

        // TODO check if hash_map is the org_keys
        let (unlocked_keys, _unlocked_org_keys) = rbw::actions::unlock(
            &self.email,
            &password,
            self.db.kdf.expect("KDF should be set"),
            self.db.iterations.expect("iterations not set"),
            self.db.memory,
            self.db.parallelism,
            self.db
                .protected_key
                .as_ref()
                .expect("Protected key not set")
                .as_str(),
            self.db
                .protected_private_key
                .as_ref()
                .expect("Protected private key not set")
                .as_str(),
            &self.db.protected_org_keys,
        )?;
        // TODO unlock and decrypt entries with the unlock key
        self.decrypt(&unlocked_keys);
        self.unlocked_keys = Some(unlocked_keys);
        self.state = VaultState::Unlocked;
        Ok(())
    }

    pub fn lock(&mut self) {
        self.unlocked_keys = None;
        self.entries = None;
        self.state = VaultState::Locked;
    }

    pub async fn sync(&mut self) {
        let client = self.get_client();
        let (protected_key, protected_private_key, _protected_org_keys, entries) = client
            .sync(
                self.db
                    .access_token
                    .as_ref()
                    .expect("access token needed for sync")
                    .as_str(),
            )
            .await
            .unwrap();

        self.db.protected_key = Some(protected_key);
        self.db.protected_private_key = Some(protected_private_key);
        // self.db.protected_org_keys = Some(protected_org_keys);
        self.db.entries = entries;
        self.db.save(&self.base_url, &self.email);
    }

    // TODO replace by db.entries?
    pub fn decrypt(&mut self, unlocked_keys: &Keys) {
        self.entries = Some(
            self.db
                .entries
                .clone()
                .iter()
                .map(|entry| decrypt(entry.clone(), unlocked_keys).expect("TODO"))
                .collect(),
        );
        // TODO part splitter
        // if let Some(entries) = &self.entries {
        //     let decrypted_entries = entries.iter().map(
        //         |entry| decrypt(entry.clone(), &unlocked_keys).expect("TODO"), // TODO
        //     );
        //     self.entries = Some(decrypted_entries.collect());
        // }
    }

    pub async fn save_to_db(&self) -> Result<(), Error> {
        self.db.save_async(&self.base_url, &self.email).await
    }

    pub async fn load_from_db(&mut self) {
        if let Ok(db) = Db::load_async(&self.base_url, &self.email).await {
            self.db = db;
        }
    }
}

impl AppModel {
    async fn switch_active_vault(&mut self, new_active_vault: VaultSetting) {
        if let Some(ref mut active_vault) = self.active_vault {
            active_vault.lock();
        }

        let new_vault = self.vaults.iter().find(|vault| {
            vault.base_url == new_active_vault.base_url && vault.email == new_active_vault.email
        });
        if let Some(new_vault) = new_vault {
            self.active_vault = Some(new_vault.to_owned());
            let save_result =
                save_active_vault_to_db(VaultSetting::from(new_vault.to_owned())).await;
        }
    }
}

#[relm4::component(async)]
impl AsyncComponent for AppModel {
    type Init = ();
    type Input = AppMsg;
    type Output = ();
    type CommandOutput = ();

    view! {
        #[root]
        main_window = &adw::Window {
            set_default_width: 400,
            set_default_height: 700,
            gtk::Box {
                set_orientation: gtk::Orientation::Vertical,

                #[local_ref]
                toast_overlay -> adw::ToastOverlay {

                    #[name = "main_stack"]
                    gtk::Stack {
                        #[transition = "SlideRight"]
                        add_child = if model.active_vault.is_none() {
                            gtk::Box {
                                set_orientation: gtk::Orientation::Vertical,

                                adw::HeaderBar {
                                    #[wrap(Some)]
                                    set_title_widget = &gtk::Box {
                                        gtk::MenuButton {
                                            // #[wrap(Some)]
                                            set_popover: Some(&gtk::PopoverMenu::from_model(Some(&main_menu))),
                                        },
                                        adw::WindowTitle {
                                            set_title: "Set up a vault"
                                        },
                                    }
                                },
                                append = &gtk::Button {
                                    set_label: "Add a vault",
                                    connect_clicked => move |_| {
                                        DIALOG_BROKER.send(VaultCreationToggleMsg::Show);
                                    }
                                },
                            }
                        } else {
                            #[name = "leaflet"]
                            adw::Leaflet {
                                set_can_navigate_back: true,
                                set_vexpand: true,
                                set_visible_child_name: if model.active_entry.is_some() { "detail" } else { "list" },

                                #[name = "list"]
                                gtk::Box {
                                    set_orientation: gtk::Orientation::Vertical,

                                    #[name = "list_headerbar"]
                                    adw::HeaderBar {
                                        #[wrap(Some)]
                                        set_title_widget = &gtk::Box {
                                            set_orientation: gtk::Orientation::Horizontal,
                                            gtk::MenuButton {
                                                // #[wrap(Some)]
                                                set_popover: Some(&gtk::PopoverMenu::from_model(Some(&main_menu))),
                                            },
                                            gtk::Button {
                                                set_icon_name: "system-search",
                                                #[watch]
                                                set_visible: if let Some(ref vault) = model.active_vault {
                                                    vault.state == VaultState::Unlocked
                                                } else { false }
                                            },
                                            adw::WindowTitle {
                                              set_title: "Bitritter"
                                            },
                                        }
                                    },
                                    append: password_input = &gtk::Box {
                                        set_orientation: gtk::Orientation::Vertical,
                                        #[watch]
                                        set_visible:  model.active_vault.as_ref().is_some_and(|acc| acc.state == VaultState::Locked),
                                        gtk::Label {
                                            #[watch]
                                            set_text: if let Some(ref vault) = model.active_vault {
                                                vault.base_url.as_str()
                                            } else { "" }
                                        },
                                        gtk::Label {
                                            #[watch]
                                            set_text: if let Some(ref vault) = model.active_vault {
                                                vault.email.as_str()
                                            } else { "" }
                                        },
                                        gtk::Label {
                                            set_text: "Enter password to unlock Vault"
                                        },
                                        #[name = "password_entry"]
                                        adw::PasswordEntryRow {
                                            set_title: "Password",
                                            set_activatable: true,
                                            connect_activate[sender] => move |password| {
                                                println!("{}", password.text());
                                                sender.input(AppMsg::UnlockActiveVault(password.text().to_string()));
                                            }
                                        },
                                        gtk::Button {
                                            set_label: "Unlock Vault",
                                            connect_clicked[sender, password_entry] => move |_| {
                                                sender.input(AppMsg::UnlockActiveVault(password_entry.text().to_string()))
                                            }
                                        }
                                    },
                                    gtk::ScrolledWindow {
                                        set_vexpand: true,
                                        #[local_ref]
                                        // This is just an anchor for our factory
                                        list_box -> gtk::ListBox {}
                                    },
                                    gtk::Box {
                                        set_orientation: gtk::Orientation::Horizontal,
                                        set_hexpand: true,

                                        gtk::Button {
                                            set_label: "All",
                                            set_hexpand: true,
                                            set_margin_horizontal: 3,
                                            connect_clicked[sender] => move |_| {
                                                sender.input(AppMsg::SetVaultItemFilter(VaultStackState::All));
                                            }
                                        },
                                        gtk::Button {
                                            set_label: "Login",
                                            set_hexpand: true,
                                            set_margin_horizontal: 3,
                                            connect_clicked[sender] => move |_| {
                                                sender.input(AppMsg::SetVaultItemFilter(VaultStackState::Login));
                                            }
                                        },
                                        gtk::Button {
                                            set_label: "Card",
                                            set_hexpand: true,
                                            set_margin_horizontal: 3,
                                            connect_clicked[sender] => move |_| {
                                                sender.input(AppMsg::SetVaultItemFilter(VaultStackState::Card));
                                            }
                                        },
                                        gtk::Button {
                                            set_label: "Identity",
                                            set_hexpand: true,
                                            set_margin_horizontal: 3,
                                            connect_clicked[sender] => move |_| {
                                                sender.input(AppMsg::SetVaultItemFilter(VaultStackState::Identity));
                                            }
                                        },
                                        gtk::Button {
                                            set_label: "Note",
                                            set_hexpand: true,
                                            set_margin_horizontal: 3,
                                            connect_clicked[sender] => move |_| {
                                                sender.input(AppMsg::SetVaultItemFilter(VaultStackState::Note));
                                            }
                                        },
                                    },
                                },
                                #[name = "detail"]
                                gtk::Box {
                                    set_orientation: gtk::Orientation::Vertical,
                                    set_hexpand: true,
                                    #[name = "detail_headerbar"]
                                    adw::HeaderBar {
                                        #[name = "back_button"]
                                        pack_start = &gtk::Button {
                                            set_icon_name: "go-previous-symbolic",
                                            connect_clicked[sender] => move |_| {
                                                sender.input(AppMsg::ShowDetailView(String::new()));
                                            },
                                        },
                                        // pack_start = &gtk::Button {
                                        //     set_icon_name: "edit",
                                        //     connect_clicked[sender] => move |_| {
                                        //         // TODO maybe we can send a message to the detail
                                        //         // component here so it can handle the edit more itself
                                        //         sender.input(AppMsg::Edit);
                                        //     }
                                        // }
                                    },
                                    #[name = "detail_anchor"]
                                    gtk::Box {
                                        set_vexpand: true,
                                        set_hexpand: true,
                                    },
                                },
                            }
                        }
                    }
                }
            }
        }
    }

    menu! {
        main_menu: {
            "Vault" {
                "Add Vault" => AddVaultAction,
                // "Switch Active Vault" => SwitchVaultAction,
            }
        }
    }

    async fn init(
        _state: Self::Init,
        root: Self::Root,
        sender: AsyncComponentSender<Self>,
    ) -> AsyncComponentParts<Self> {
        tokio::time::sleep(Duration::from_millis(1)).await;

        let items = AsyncFactoryVecDeque::builder().launch_default().forward(
            sender.input_sender(),
            |output| match output {
                ListOutput::Select(id) => AppMsg::ShowDetailView(id),
            },
        );

        let init_db_result = init_db().await;

        let mut vaults: Vec<Vault> = Vec::new();
        match load_vault_settings_from_db().await {
            Ok(vault_settings) => {
                vaults = vault_settings
                    .into_iter()
                    .map(|setting| {
                        Vault::new(
                            setting.base_url,
                            setting.identity_url.unwrap_or_default(),
                            None,
                            setting.email,
                        )
                    })
                    .collect();
            }
            Err(e) => println!("{:?}", e),
        }

        let toaster = Toaster::default();

        // TODO implement better error handling here
        let mut loaded_active_vault = load_active_vault_from_db().await.unwrap_or(None);
        if let Some(ref mut vault) = loaded_active_vault {
            vault.load_from_db().await;
        }

        let dialog = VaultCreationModel::builder()
            .transient_for(&root)
            .launch_with_broker((), &DIALOG_BROKER)
            .forward(sender.input_sender(), |msg| msg);

        let model = AppModel {
            vault_stack_state: VaultStackState::All,
            vaults,
            active_vault: loaded_active_vault,
            active_entry: None,
            items,
            toaster,
            dialog,
        };

        let toast_overlay = model.toaster.overlay_widget();

        let list_box = model.items.widget();
        // list_box.set_filter_func(|item| match item.downcast::<ListItem>() {
        //     Ok(list_item) => list_item.item_type == VaultItemType::Login
        //     Err(_) => todo!(),
        // });

        let widgets = view_output!();

        let vault_add_action: RelmAction<AddVaultAction> = RelmAction::new_stateless(move |_| {
            sender.input(AppMsg::SetupVault);
        });

        let mut vault_action_group = RelmActionGroup::<VaultActionGroup>::new();

        vault_action_group.add_action(vault_add_action);
        vault_action_group.register_for_widget(&widgets.main_window);

        // Bindings for the Leaflet interaction
        widgets
            .leaflet
            .bind_property("folded", &widgets.list_headerbar, "show-end-title-buttons")
            .flags(glib::BindingFlags::SYNC_CREATE)
            .build();

        widgets
            .leaflet
            .bind_property(
                "folded",
                &widgets.detail_headerbar,
                "show-start-title-buttons",
            )
            .flags(glib::BindingFlags::SYNC_CREATE)
            .build();

        widgets
            .leaflet
            .bind_property("folded", &widgets.back_button, "visible")
            .flags(glib::BindingFlags::SYNC_CREATE)
            .build();

        AsyncComponentParts { model, widgets }
    }

    async fn update_with_view(
        &mut self,
        widgets: &mut AppModelWidgets,
        message: Self::Input,
        sender: AsyncComponentSender<Self>,
        _root: &Self::Root,
    ) {
        tokio::time::sleep(Duration::from_millis(1)).await;

        match message {
            AppMsg::Startup => {
                if let Some(ref mut vault) = self.active_vault {
                    vault.load_from_db().await;
                }
            }
            AppMsg::Login(password) => {
                if let Some(ref mut vault) = self.active_vault {
                    vault.pre_login().await;
                    if (vault.login(password.as_str()).await).is_ok() {}
                    println!("Logged in");
                }
            }
            AppMsg::Sync => {
                if let Some(ref mut vault) = self.active_vault {
                    vault.sync().await;
                    println!("Sync complete");
                }
            }
            AppMsg::UnlockActiveVault(password) => {
                if let Some(ref mut vault) = self.active_vault {
                    vault.load_from_db().await;
                    let mut toast_title: String = "Unlock Error".to_owned();
                    let mut login_complete: bool = false;
                    if !vault.is_logged_in() {
                        match vault.pre_login().await {
                            Ok(_) => match vault.login(password.as_str()).await {
                                Err(login_error) => {
                                    match login_error {
                                        Error::ConfigMissingEmail => {
                                            toast_title = "Config missing E-Mail".to_owned()
                                        }
                                        Error::IncorrectApiKey => {
                                            toast_title = "Incorrect API Key".to_owned()
                                        }
                                        Error::IncorrectPassword { message } => {
                                            toast_title = message
                                        }
                                        _ => toast_title = "Undefined Login Error".to_owned(),
                                    }
                                    let error_toast =
                                        Toast::builder().title(&toast_title).timeout(5).build();
                                    widgets.toast_overlay.add_toast(error_toast);
                                }
                                Ok(_) => {
                                    vault.sync().await;
                                    login_complete = true;
                                }
                            },
                            Err(pre_login_error) => match pre_login_error {
                                Error::ConfigMissingEmail => {
                                    toast_title = "Config missing E-Mail".to_owned()
                                }
                                Error::IncorrectApiKey => {
                                    toast_title = "Incorrect API Key".to_owned()
                                }
                                _ => toast_title = "Undefined Pre-Login Error".to_owned(),
                            },
                        }
                    } else {
                        login_complete = true;
                    }
                    if login_complete {
                        match vault.unlock(password.as_str()) {
                            Err(unlock_error) => {
                                match unlock_error {
                                    Error::IncorrectPassword { message } => toast_title = message,
                                    _ => {}
                                }
                                let error_toast =
                                    Toast::builder().title(&toast_title).timeout(5).build();
                                widgets.toast_overlay.add_toast(error_toast);
                            }
                            Ok(_) => {
                                // TODO i suspect the to_owned() clone makes the loading that slow
                                for entry in vault.entries.to_owned().unwrap_or_else(Vec::new) {
                                    self.items.guard().push_front(ListItem::from(entry));
                                }
                                println!("Unlock complete");
                            }
                        }
                        self.update_view(widgets, sender);
                    }
                } else {
                    // we don't have an active vault here
                    // maybe spawn a toast just in case?
                }
            }
            AppMsg::LockActiveVault => {
                if let Some(ref mut vault) = self.active_vault {
                    vault.lock();
                    self.update_view(widgets, sender);
                }
            }

            // AppMsg::AddVaultItem => {}
            AppMsg::SetupVault => {
                DIALOG_BROKER.send(VaultCreationToggleMsg::Show);
            }
            AppMsg::SwitchVault(new_vault) => {
                self.switch_active_vault(new_vault).await;
                self.update_view(widgets, sender);
            }
            AppMsg::SetVaultItemFilter(filter) => {
                self.vault_stack_state = filter;
            }
            AppMsg::CreateVault(vault_setting) => {
                let new_vault = Vault::from(vault_setting.clone());
                self.vaults.push(new_vault);

                match insert_vault_setting_to_db(vault_setting.clone()).await {
                    Ok(_) => {
                        DIALOG_BROKER.send(VaultCreationToggleMsg::Hide);
                        let success_toast =
                            Toast::builder().title("Vault Added").timeout(5).build();
                        widgets.toast_overlay.add_toast(success_toast);
                    }
                    Err(_) => {
                        let success_toast = Toast::builder()
                            .title("Error saving vault settings to database")
                            .timeout(5)
                            .build();
                        widgets.toast_overlay.add_toast(success_toast);
                    }
                }
                if vault_setting.set_active {
                    self.switch_active_vault(VaultSetting::from(vault_setting))
                        .await;
                }
                self.update_view(widgets, sender);
            }
            AppMsg::ShowDetailView(id) => {
                if id.is_empty() {
                    widgets.leaflet.navigate(adw::NavigationDirection::Back);
                } else {
                    widgets.leaflet.navigate(adw::NavigationDirection::Forward);

                    // TODO prettify those unwraps
                    let new_entry = self
                        .active_vault
                        .as_ref()
                        .unwrap()
                        .entries
                        .as_ref()
                        .unwrap()
                        .clone()
                        .into_iter()
                        .find(|entry| entry.id == id)
                        .unwrap()
                        .clone();
                    self.active_entry = Some(new_entry.clone());

                    if let Some(ref detail_widget) = widgets.detail_anchor.first_child() {
                        widgets.detail_anchor.remove(detail_widget);
                    }
                    // TODO maybe we can refactor code here to not move the full entry there,
                    // but rather just the DetailLogin or DetailCard structures
                    match new_entry.data {
                        EntryData::Login { .. } => {
                            let detail_login = DetailLogin::builder()
                                .launch(new_entry)
                                .forward(sender.input_sender(), detail_msg_to_app_msg_transformer);
                            widgets.detail_anchor.append(detail_login.widget());
                        }
                        EntryData::Card { .. } => {
                            let detail_card = DetailCard::builder()
                                .launch(new_entry)
                                .forward(sender.input_sender(), detail_msg_to_app_msg_transformer);
                            widgets.detail_anchor.append(detail_card.widget());
                        }
                        EntryData::Identity { .. } => todo!(),
                        EntryData::SecureNote => todo!(),
                    };
                    self.update_view(widgets, sender);
                }
            }
            AppMsg::ClipboardCopy(string_to_copy) => {
                // Sadly we have to use 2 crates to copy
                // This time for X11 / Win / MacOS
                if let Ok(mut ctx) = ClipboardContext::new() {
                    if let Ok(()) = ctx.set_contents(string_to_copy.clone()) {
                        // We do this just because it mostly does not work to only set the content
                        if ctx.get_contents().is_ok() {}
                        println!("Spawning toast");
                        let toast = Toast::builder()
                            .title("Value copied to clipboard")
                            .timeout(5)
                            .build();
                        self.toaster.add_toast(toast);
                    } else {
                        // TODO error handling, though it might be quite fatal at this point
                    }
                } else {
                    // TODO error handling, though it might be quite fatal at this point
                }
                // This time for wayland
                let opts = Options::new();
                match opts.copy(
                    Source::Bytes(string_to_copy.into_bytes().into()),
                    MimeType::Autodetect,
                ) {
                    Ok(_) => {
                        let toast = Toast::builder()
                            .title("Value copied to clipboard")
                            .timeout(5)
                            .build();
                        self.toaster.add_toast(toast);
                    }
                    Err(_) => {
                        // TODO error handling, though it might be quite fatal at this point
                    }
                }
            }
            AppMsg::Edit => {
                // TODO we need to update the widgets of the detail_box
                self.update_view(widgets, sender);
            }
            AppMsg::Discard => {}
        }
    }
}

fn detail_msg_to_app_msg_transformer(msg: DetailMsg) -> AppMsg {
    match msg {
        DetailMsg::Edit => AppMsg::Edit,
        DetailMsg::Discard => {
            println!("Discard");
            AppMsg::Discard
        }
        DetailMsg::ClipboardCopy(string_to_copy) => AppMsg::ClipboardCopy(string_to_copy),
    }
}

relm4::new_action_group!(VaultActionGroup, "vault");
relm4::new_stateless_action!(AddVaultAction, VaultActionGroup, "add_vault");
relm4::new_stateless_action!(SwitchVaultAction, VaultActionGroup, "switch_vault");

fn main() {
    gtk::init().unwrap();
    let app = RelmApp::new("dev.bitritter");
    relm4_icons::initialize_icons();
    app.run_async::<AppModel>(());
}
